@@ -114,27 +114,79 @@ extern "C" {
      0 for n data, n children
      */
     
-#define CNEARTREE_FLAG_LEFT_DATA   1          /* 0x0001 */
-#define CNEARTREE_FLAG_RIGHT_DATA  2          /* 0x0002 */
-#define CNEARTREE_FLAG_LEFT_CHILD  4          /* 0x0004 */
-#define CNEARTREE_FLAG_RIGHT_CHILD 8          /* 0x0008 */
+#define CNEARTREE_FLAG_LEFT_DATA   1L          /* 0x0001 */
+#define CNEARTREE_FLAG_RIGHT_DATA  2L          /* 0x0002 */
+#define CNEARTREE_FLAG_LEFT_CHILD  4L          /* 0x0004 */
+#define CNEARTREE_FLAG_RIGHT_CHILD 8L          /* 0x0008 */
     
-#define CNEARTREE_DATA_OR_CHILDREN 15         /* 0x000F */
+#define CNEARTREE_DATA_OR_CHILDREN 15L         /* 0x000F */
     
-#define CNEARTREE_TYPE_DOUBLE      16         /* 0x0010 */
-#define CNEARTREE_TYPE_INTEGER     32         /* 0x0020 */
+#define CNEARTREE_TYPE_DOUBLE      16L         /* 0x0010 */
+#define CNEARTREE_TYPE_INTEGER     32L         /* 0x0020 */
+#define CNEARTREE_TYPE_STRING      64L         /* 0x0040 */
     
-#define CNEARTREE_TYPE             48         /* 0x0030 */
+#define CNEARTREE_TYPE            112L         /* 0x0070 */
     
-#define CNEARTREE_NORM_UNKNOWN     64         /* 0x0040 */
-#define CNEARTREE_NORM_L1         128         /* 0x0080 */
-#define CNEARTREE_NORM_L2         256         /* 0x0100 */
-#define CNEARTREE_NORM_LINF       512         /* 0x0200 */
+#define CNEARTREE_NORM_UNKNOWN    128L         /* 0x0080 */
+#define CNEARTREE_NORM_L1         256L         /* 0x0100 */
+#define CNEARTREE_NORM_L2         512L         /* 0x0200 */
+#define CNEARTREE_NORM_LINF      1024L         /* 0x0400 */
+#define CNEARTREE_NORM_SPHERE    2048L         /* 0x0800 */
+#define CNEARTREE_NORM_HSPHERE   4096L         /* 0x1000 */
+#define CNEARTREE_NORM_HAMMING   8192L         /* 0x2000 */
+#define CNEARTREE_NORM_L2LAZY   16384L         /* 0x4000 */
     
-#define CNEARTREE_NORM            960         /* 0x03C0 */
+#define CNEARTREE_NORM          32640L         /* 0x7F80 */
     
-#define CNEARTREE_FLIP           1024         /* 0x0400 */
-#define CNEARTREE_DEFER_ALL      2048         /* 0x0800 */
+    
+/*  Execution Control Flags */
+    
+#define CNTF_NOPREPRUNE    0x10000L     /*flag to supress all search prepruning */
+#define CNTF_FORCEPREPRUNE 0x20000L     /*flag to force search prepruning       */
+#define CNTF_NOFLIP        0x40000L     /*flag to suppress flips on insert      */
+#define CNTF_FORCEFLIP     0x80000L     /*flag to force flips on insert         */
+#define CNTF_NODEFER      0x100000L     /*flag to prevent deferred insert       */
+
+    
+#define CNEARTREE_XFLAGS  0x1F0000L     /*mask for execution flags */
+    
+#ifdef CNEARTREE_FORCEPREPRUNE
+  #define CNFT_FLAGDEFAULTPRUNE  CNTF_FORCEPREPRUNE
+  #ifdef CNEARTREE_NOPREPRUNE
+    #error "CNEARTREE_NOPREPRUNE conflicts with  CNEARTREE_FORCEPREPRUNE"
+  #endif
+#else
+  #ifdef CNEARTREE_NOPREPRUNE
+    #define  CNFT_FLAGDEFAULTPRUNE  CNTF_NOPREPRUNE
+  #else
+    #define  CNFT_FLAGDEFAULTPRUNE  0
+  #endif
+#endif
+    
+#ifdef CNEARTREE_FORCEFLIP
+  #define CNFT_FLAGDEFAULTFLIP  CNTF_FORCEFLIP
+  #ifdef CNEARTREE_NOFLIP
+    #error "CNEARTREE_NOFLIP conflicts with  CNEARTREE_FORCEFLIP"
+  #endif
+#else
+  #ifdef CNEARTREE_NOFLIP
+    #define CNFT_FLAGDEFAULTFLIP CNTF_NOFLIP
+  #else
+    #define CNFT_FLAGDEFAULTFLIP 0
+  #endif
+#endif
+    
+#ifdef CNEARTREE_NODEFER
+    #define CNFT_FLAGDEFAULTDEFER CNTF_NODEFER
+#else
+    #define CNFT_FLAGDEFAULTDEFER 0
+#endif
+    
+#define CNTF_FLAGSDEFAULT  (CNFT_FLAGDEFAULTPRUNE|CNFT_FLAGDEFAULTFLIP|CNFT_FLAGDEFAULTDEFER)
+    
+    
+#define CNEARTREE_FLIP          0     /* Not used, defined for compatibility */
+#define CNEARTREE_DEFER_ALL     0     /* Not used, defined for compatibility */
     
     
     
@@ -151,7 +203,7 @@ extern "C" {
         /* tree descending from the left object        */
         struct _CNearTreeNode CNEARTREE_FAR * m_pRightBranch; 
         /* tree descending from the right object       */
-        int              m_iflags;       /* flags                                       */
+        long              m_iflags;       /* flags      */
     } CNearTreeNode;
     
     
@@ -162,17 +214,141 @@ extern "C" {
         size_t           m_szdimension;   /* dimension of the coordinates                */
         size_t           m_szsize;        /* size of this tree                           */
         size_t           m_szdepth;       /* depth of this tree                          */
-        int              m_iflags;        /* flags                                       */
+        long             m_iflags;        /* flags                                       */
         CVectorHandle    m_ObjectStore;   /* all inserted objects                        */
         CVectorHandle    m_CoordStore;    /* all inserted coordinates                    */
         CVectorHandle    m_DelayedIndices;/* objects queued for insertion                */
 
         CRHrand          m_rhr;           /* random number generator                     */
-
+        double           m_DiamEstimate;  /* estimated diameter */
+        double           m_SumSpacings;   /* sum of spacings at time of insertion */
+        double           m_SumSpacingsSq; /* sum of spacings squared at time of insertion */
+        double           m_DimEstimate;   /* estimated dimension */
+        double           m_DimEstimateEsd;/* estimated dimension estimated standard deviation */
+#ifdef CNEARTREE_INSTRUMENTED
+        size_t           m_NodeVisits;    /* number of node visits */
+#endif
     } CNearTree;
     
     typedef CNearTree     CNEARTREE_FAR * CNearTreeHandle;
     
+/* Distance Macros for L2LAZY */
+    
+    
+#define CNTM_Abs(x) (((x)<0.)?(-(x)):(x))
+#define CNTM_DistL2sq(distsq,treehandle,coord1,coord2) {     \
+        size_t index;                                     \
+        size_t treedim; \
+        long treetype;  \
+        double CNEARTREE_FAR * dcoord1 = NULL; \
+        double CNEARTREE_FAR * dcoord2 = NULL; \
+        int CNEARTREE_FAR * icoord1 = NULL; \
+        int CNEARTREE_FAR * icoord2 = NULL; \
+        treedim = (treehandle)->m_szdimension; \
+        treetype = (treehandle)->m_iflags&CNEARTREE_TYPE; \
+        if (treetype == CNEARTREE_TYPE_DOUBLE) { \
+            dcoord1 = (double CNEARTREE_FAR *)(coord1);\
+            dcoord2 = (double CNEARTREE_FAR *)(coord2);\
+        } else if (treetype == CNEARTREE_TYPE_INTEGER) { \
+            icoord1 = (int CNEARTREE_FAR *)(coord1); \
+            icoord2 = (int CNEARTREE_FAR *)(coord2); \
+        } \
+        distsq = 0.; \
+        if (treetype == CNEARTREE_TYPE_DOUBLE) { \
+            distsq = (dcoord1[0]-dcoord2[0])*(dcoord1[0]-dcoord2[0]); \
+            \
+            for (index=1; index < treedim; index++) { \
+                distsq += (dcoord1[index]-dcoord2[index]) \
+                *(dcoord1[index]-dcoord2[index]); \
+            } \
+        } else if (treetype == CNEARTREE_TYPE_INTEGER) { \
+            distsq = ((double)(icoord1[0]-icoord2[0]))*((double)(icoord1[0]-icoord2[0])); \
+            \
+            for (index=1; index < treedim; index++) { \
+                distsq += ((double)(icoord1[index]-icoord2[index])) \
+                *((double)(icoord1[index]-icoord2[index])); \
+            } \
+        }  \
+        }
+
+#define CNTM_DistL2(dist,treehandle,coord1,coord2) {     \
+        size_t index;                                     \
+        double distsq=0.;                                    \
+        size_t treedim; \
+        long treetype;  \
+        double CNEARTREE_FAR * dcoord1 = NULL; \
+        double CNEARTREE_FAR * dcoord2 = NULL; \
+        int CNEARTREE_FAR * icoord1 = NULL; \
+        int CNEARTREE_FAR * icoord2 = NULL; \
+        treedim = (treehandle)->m_szdimension; \
+        treetype = (treehandle)->m_iflags&CNEARTREE_TYPE; \
+        if (treetype == CNEARTREE_TYPE_DOUBLE) { \
+            dcoord1 = (double CNEARTREE_FAR *)(coord1);\
+            dcoord2 = (double CNEARTREE_FAR *)(coord2);\
+        } else if (treetype == CNEARTREE_TYPE_INTEGER) { \
+            icoord1 = (int CNEARTREE_FAR *)(coord1); \
+            icoord2 = (int CNEARTREE_FAR *)(coord2); \
+        } \
+        if (treedim == 1) { \
+            if (treetype == CNEARTREE_TYPE_DOUBLE) { \
+                (dist) = CNTM_Abs(dcoord1[0]-dcoord2[0]); \
+            } else if (treetype == CNEARTREE_TYPE_INTEGER) { \
+                (dist) = CNTM_Abs((double)(icoord1[0]-icoord2[0])); \
+            } \
+        } else { \
+            if (treetype == CNEARTREE_TYPE_DOUBLE) { \
+                distsq = (dcoord1[0]-dcoord2[0])*(dcoord1[0]-dcoord2[0]); \
+                \
+                for (index=1; index < treedim; index++) { \
+                    distsq += (dcoord1[index]-dcoord2[index]) \
+                    *(dcoord1[index]-dcoord2[index]); \
+                } \
+            } else if (treetype == CNEARTREE_TYPE_INTEGER) { \
+                distsq = ((double)(icoord1[0]-icoord2[0]))*((double)(icoord1[0]-icoord2[0])); \
+                \
+                for (index=1; index < treedim; index++) { \
+                    distsq += ((double)(icoord1[index]-icoord2[index])) \
+                    *((double)(icoord1[index]-icoord2[index])); \
+                } \
+            }  \
+            (dist) = sqrt(distsq);\
+        }\
+        }
+  
+#define CNTM_DistL1(dist,treehandle,coord1,coord2) {\
+        size_t index;   \
+        size_t treedim; \
+        long treetype;  \
+        double CNEARTREE_FAR * dcoord1 = NULL; \
+        double CNEARTREE_FAR * dcoord2 = NULL; \
+        int CNEARTREE_FAR * icoord1 = NULL; \
+        int CNEARTREE_FAR * icoord2 = NULL; \
+        \
+        treedim = treehandle->m_szdimension; \
+        treetype = treehandle->m_iflags&CNEARTREE_TYPE; \
+        (dist) = 0.; \
+        \
+        if (treetype == CNEARTREE_TYPE_DOUBLE) { \
+            dcoord1 = (double CNEARTREE_FAR *)(coord1);\
+            dcoord2 = (double CNEARTREE_FAR *)(coord2);\
+        } else if (treetype == CNEARTREE_TYPE_INTEGER) { \
+            icoord1 = (int CNEARTREE_FAR *)(coord1); \
+            icoord2 = (int CNEARTREE_FAR *)(coord2); \
+        } \
+       if (treetype == CNEARTREE_TYPE_DOUBLE) { \
+            (dist)= fabs(dcoord1[0]-dcoord2[0]); \
+            for (index=1; index < treedim; index++) { \
+                (dist) += CNTM_Abs(dcoord1[index]-dcoord2[index]); \
+            } \
+        } else if (treetype == CNEARTREE_TYPE_INTEGER) { \
+            (dist) = fabs((double)(icoord1[0]-icoord2[0])); \
+            for (index=1; index < treedim; index++) { \
+                (dist) += CNTM_Abs((double)(dcoord1[index]-dcoord2[index])); \
+            } \
+        }   \
+    }
+
+    
     /*
      =======================================================================
      double CNearTreeDistsq(void CNEARTREE_FAR * coord1, 
@@ -197,20 +373,24 @@ extern "C" {
     double CNearTreeDistsq(void CNEARTREE_FAR * coord1,
                            void CNEARTREE_FAR * coord2, 
                            size_t treedim, 
-                           int treetype);
-    /*
+                           long treetype);
+
+     /*
      =======================================================================
-     double CNearTreeDist(const CNearTreeHandle treehandle, void CNEARTREE_FAR * coord1, 
+     double CNearTreeDist(const CNearTreeHandle treehandle, 
+     void CNEARTREE_FAR * coord1, 
      void CNEARTREE_FAR * coord2)
      
      function to return the distance (L1, L2 or L-infinity) between two 
      coordinate vectors according to the parameters of the given tree  
+     For L2LAZY tree this returns the L2 norm distance.
      
      =======================================================================
      */
-    double CNearTreeDist(const CNearTreeHandle treehandle, void CNEARTREE_FAR * coord1, 
-                         void CNEARTREE_FAR * coord2);
     
+    double CNearTreeDist(const CNearTreeHandle treehandle, 
+                              void CNEARTREE_FAR * coord1,
+                              void CNEARTREE_FAR * coord2);
     /*
      =======================================================================
      int CNearTreeSetNorm(const CNearTreeHandle treehandle, int treenorm);
@@ -220,6 +400,7 @@ extern "C" {
      treenorm should be one of CNEARTREE_NORM_L1 for an L-1 norm
      CNEARTREE_NORM_L2 for an L-2 norm
      CNEARTREE_NORM_LINF for an L-infinity norm
+     CNEARTREE_NORM_LAZY for an L-1 norm used for L-2 searching
      
      the function returns CNEARTREE_BAD_ARGUMENT for an invalid argument
      CNEARTREE_SUCCESS (0) otherwise
@@ -255,18 +436,33 @@ extern "C" {
     /*
      =======================================================================
      int CNearTreeCreate ( CNearTreeHandle CNEARTREE_FAR * treehandle, 
-     size_t treedim, 
-     int treetype )
+     size_t treedim, long treetype)
      
      Create a CNearTree
      
-     treedim -- the dimension of the vectors
-     treetype -- and integer flag for type of the vectors
-     CNEARTREE_TYPE_DOUBLE for double
-     CNEARTREE_TYPE_INTEGER for integer     
-     
      returns a pointer to the newly allocated block of memory as a 
-     CNearTreeHandle in *treehandle, assuming coordinates of dimension treedim
+     CNearTreeHandle in *treehandle
+     
+     
+     treedim -- the dimension of the vectors
+     treetype -- double or integer flag for type of the vectors ored with norm
+                 and ored with execution flags
+     CNEARTREE_TYPE_DOUBLE for double
+     CNEARTREE_TYPE_INTEGER for integer
+     CNEARTREE_TYPE_STRING for strings
+     ored with
+     CNEARTREE_NORM_L1        for the sum of the absolute values
+     CNEARTREE_NORM_L2        for the square root of the sum of the squares
+     CNEARTREE_NORM_LINF      for the max
+     CNEARTREE_NORM_SPHERE    for norm as spherical angular distance
+     CNEARTREE_NORM_HSPHERE   for norm as hemispherical angular distance
+     CNEARTREE_NORM_HAMMING   for norm as string hamming distance
+     ored with
+     CNTF_NOPREPRUNE    0x10000L    flag to supress all search prepruning
+     CNTF_FORCEPREPRUNE 0x20000L    flag to force search prepruning
+     CNTF_NOFLIP        0x40000L    flag to suppress flips on insert
+     CNTF_FORCEFLIP     0x80000L    flag to force flips on insert
+     CNTF_NODEFER      0x100000L    flag to prevent deferred insert 
      
      creates an empty tree with no right or left node and with the dMax-below
      set to negative values so that any match found will be stored since it will
@@ -277,7 +473,7 @@ extern "C" {
     
     int CNearTreeCreate(CNearTreeHandle CNEARTREE_FAR * treehandle, 
                         size_t treedim, 
-                        int treetype);
+                        long treetype);
     
     /*
      =======================================================================
@@ -337,11 +533,23 @@ extern "C" {
      int CNearTreeGetDelayedSize (const CNearTreeHandle treehandle, size_t CNEARTREE_FAR * size)
      
      Return the number of objects in the delay queue tree in size
+     This is a deprecated alternate name for CNearTreeGetDeferredSize
      
      =======================================================================
      */
     
     int CNearTreeGetDelayedSize (const CNearTreeHandle treehandle, size_t CNEARTREE_FAR * size);
+
+    /*
+     =======================================================================
+     int CNearTreeGetDeferredSize (const CNearTreeHandle treehandle, size_t CNEARTREE_FAR * size)
+     
+     Return the number of objects in the delay queue tree in size
+     
+     =======================================================================
+     */
+    
+    int CNearTreeGetDeferredSize ( const CNearTreeHandle treehandle, size_t CNEARTREE_FAR * size );
     
     /*
      =======================================================================
@@ -364,7 +572,174 @@ extern "C" {
      */
     
     int CNearTreeGetDepth (const CNearTreeHandle treehandle, size_t CNEARTREE_FAR * depth);
+
     
+    /*
+     =======================================================================
+
+     int CNearTreeGetFlags(const CNearTreeHandle treehandle, 
+     long CNEARTREE_FAR *flags, 
+     const long mask)
+     
+     int CNearTreeSetFlags(const CNearTreeHandle treehandle,
+     const long flags,
+     const long mask)
+     
+     functions to get or set the execution control flags:
+     
+        CNTF_NOPREPRUNE    to supress all search prepruning
+        CNTF_FORCEPREPRUNE to force search prepruning      
+        CNTF_NOFLIP        flag to suppress flips on insert
+        CNTF_FORCEFLIP     flag to force flips on insert
+        CNTF_NODEFER       flag to prevent deferred insert
+    
+     The desired flags may be ored.  If mask is no zero it is
+     used as a bit mask, so a call with a flags of zero and a
+     mask equal to a particular flag, clears that flag.
+     =======================================================================
+     */
+    
+    int CNearTreeGetFlags(const CNearTreeHandle treehandle, 
+                          long CNEARTREE_FAR *flags, 
+                          const long mask);
+    
+    int CNearTreeSetFlags(const CNearTreeHandle treehandle,
+                          const long flags,
+                          const long mask);
+    
+    /*
+     =======================================================================
+     int CNearTreeGetMeanSpacing ( const CNearTreeHandle treehandle, 
+     double CNEARTREE_FAR * spacing  )
+     
+     Get an estimate of the spacing of points
+     
+     
+     =======================================================================
+     */
+    
+    int CNearTreeGetMeanSpacing ( const CNearTreeHandle treehandle, 
+                                 double CNEARTREE_FAR * spacing  );
+    
+    /*
+     =======================================================================
+     int CNearTreeGetVarSpacing ( const CNearTreeHandle treehandle, 
+     double CNEARTREE_FAR * varspacing  )
+     
+     Get an estimate of variance of the spacing of points
+     
+     
+     =======================================================================
+     */
+    int CNearTreeGetVarSpacing ( const CNearTreeHandle treehandle, 
+                                double CNEARTREE_FAR * varspacing  );
+    
+    /*
+     =======================================================================
+     int CNearTreeCount(const CNearTreeHandle treehandle, 
+     size_t CNEARTREE_FAR * count)
+     =======================================================================
+     */
+    
+    int CNearTreeCount(const CNearTreeHandle treehandle, 
+                       size_t CNEARTREE_FAR * count);
+                        
+    
+#ifndef CNEARTREE_INSTRUMENTED
+    /*
+     =======================================================================
+     int CNearTreeGetNodeVisits (  const CNearTreeHandle treehandle,
+     size_t CNEARTREE_FAR * visits)
+     
+     Get the number of visits to nodes
+     Dummy version to return 0
+     
+     
+     =======================================================================
+     */
+    int CNearTreeGetNodeVisits ( const CNearTreeHandle treehandle,
+                                size_t CNEARTREE_FAR * visits);
+    
+#endif
+    
+#ifdef CNEARTREE_INSTRUMENTED
+    /*
+     =======================================================================
+     int CNearTreeGetNodeVisits ( const CNearTreeHandle treehandle,
+     size_t CNEARTREE_FAR * visits)
+     
+     Get the number of visits to nodes
+     
+     
+     =======================================================================
+     */
+    int CNearTreeGetNodeVisits ( const CNearTreeHandle treehandle,
+                                size_t CNEARTREE_FAR * visits);
+    /*
+     =======================================================================
+     int CNearTreeSetNodeVisits (  const CNearTreeHandle treehandle,
+     const size_t visits )
+     
+     Set the number of visits to nodes
+     
+     
+     =======================================================================
+     */
+    int CNearTreeSetNodeVisits (  const CNearTreeHandle treehandle,
+                                const size_t visits );
+#endif    
+    
+    /*
+     =======================================================================
+     int CNearTreeGetDiamEstimate (  const CNearTreeHandle treehandle,
+     double CNEARTREE_FAR * diamest )
+     
+     Get an estimate of the diameter
+     
+     
+     =======================================================================
+     */
+    int CNearTreeGetDiamEstimate ( const CNearTreeHandle treehandle,
+                                  double CNEARTREE_FAR * diamest );
+    
+    /*
+     =======================================================================
+     int CNearTreeGetDimEstimateEsd ( const CNearTreeHandle treehandle,
+     double CNEARTREE_FAR * dimestesd )
+     
+     Get the current best estimate of the dimension esd
+     
+     =======================================================================
+     */
+    int CNearTreeGetDimEstimateEsd ( const CNearTreeHandle treehandle,
+                                    double CNEARTREE_FAR * dimestesd );
+    
+    
+    /*
+     =======================================================================
+     int CNearTreeGetDimEstimate ( const CNearTreeHandle treehandle,
+     double CNEARTREE_FAR * diamest,
+     const double DimEstimateEsd )
+     
+     Get an estimate of the dimension of the collection of points
+     in the tree, to within the specified esd
+     
+     
+     =======================================================================
+     */
+    int CNearTreeGetDimEstimate ( const CNearTreeHandle treehandle,
+                                 double CNEARTREE_FAR * dimest,
+                                 const double DimEstimateEsd );
+    
+    /*
+     =======================================================================
+    int CNearTreeNodeCount(const CNearTreeNodeHandle treenodehandle, 
+                           size_t CNEARTREE_FAR * count)
+     =======================================================================
+     */
+    
+    int CNearTreeNodeCount(const CNearTreeNodeHandle treenodehandle, 
+                           size_t CNEARTREE_FAR * count);
     /*
      =======================================================================
      int CNearTreeImmediateInsert ( const CNearTreeHandle treehandle, 
@@ -393,19 +768,53 @@ extern "C" {
                         const void CNEARTREE_FAR * coord, 
                         const void CNEARTREE_FAR * obj );
     
+    /*  
+     =======================================================================
+     int CNearTreeNodeReInsert_Flip ( const CNearTreeHandle treehandle,
+     const const CNearTreeNodeHandle treenodehandle, 
+     size_t CNEARTREE_FAR * depth)
+     
+     Function to reinsert the elements from a detached a node into a CNearTree 
+     for later searching
+     
+     treehandle is the handle of the overall neartree being used
+     
+     treenodehandle is the handle of the node in the existing tree at which to start
+     the insertion
+     
+     pntn is the handle of the previously detached node from which to 
+     extract the nodes for insertion
+     
+     depth is used to keep track of the depth at which the insertion is done
+     
+     return 0 for success, nonzero for an error
+     
+     =======================================================================
+     */
+    
+    int CNearTreeNodeReInsert_Flip( const CNearTreeHandle treehandle,
+                                   const CNearTreeNodeHandle treenodehandle,
+                                   const CNearTreeNodeHandle pntn,
+                                   size_t CNEARTREE_FAR * depth);
+        
+        
+    
     
     /*
      =======================================================================
-     int CNearTreeNodeInsert( const CNearTreeHandle treehandle,
-     const CNearTreeNodeHandle treenodehandle,
-     const size_t index,
-     size_t CNEARTREE_FAR * depth );
+     int CNearTreeNodeInsert ( const CNearTreeHandle treehandle,
+     const const CNearTreeNodeHandle treenodehandle, 
+     size_t index;
+     size_t CNEARTREE_FAR * depth)
      
-     Function to insert some "point" as an object into a CNearTree for
-     later searching, starting at a given treenode
+     Function to insert some "point" as an object into a node in a CNearTree for
+     later searching
      
-     coord is a coordinate vector for an object, obj, to be inserted into a
-     Neartree.  A static copy of the coordinates and a pointer to the
+     treenodehandle is the handle of the node at which to start the insertion
+     
+     index is the index of the object and coordinates to add from 
+     treehandle->m_ObjectStore and treehandle->CoordStore into a NearTree.
+     A static copy of the coordinates and a pointer to the
      object are inserted
      
      Three possibilities exist: put the datum into the left
@@ -418,13 +827,42 @@ extern "C" {
      return 0 for success, nonzero for an error
      
      =======================================================================
-     */
+     =======================================================================
+     int CNearTreeNodeInsert_Flip ( const CNearTreeHandle treehandle,
+     const const CNearTreeNodeHandle treenodehandle, 
+     size_t index;
+     size_t CNEARTREE_FAR * depth)
+     
+     Function to insert some "point" as an object into a node in a CNearTree for
+     later searching
+     
+     treenodehandle is the handle of the node at which to start the insertion
+     
+     index is the index of the object and coordinates to add from 
+     treehandle->m_ObjectStore and treehandle->CoordStore into a NearTree.
+     A static copy of the coordinates and a pointer to the
+     object are inserted
+     
+     Three possibilities exist: put the datum into the left
+     position (first test),into the right position, or else
+     into a node descending from the nearer of those positions
+     when they are both already used.
+     
+     depth is used to keep track of the depth at which the insertion is done
+     
+     return 0 for success, nonzero for an error
+     
+     =======================================================================
+        */
     
     int CNearTreeNodeInsert( const CNearTreeHandle treehandle,
                             const CNearTreeNodeHandle treenodehandle,
                             const size_t index, 
                             size_t CNEARTREE_FAR * depth );
-    
+    int CNearTreeNodeInsert_Flip( const CNearTreeHandle treehandle,
+                            const CNearTreeNodeHandle treenodehandle,
+                            const size_t index, 
+                            size_t CNEARTREE_FAR * depth );    
     
     /*
      =======================================================================
@@ -448,6 +886,7 @@ extern "C" {
      return 0 for success, nonzero for an error
      
      =======================================================================
+     
      */
     
     int CNearTreeInsert( const CNearTreeHandle treehandle,
@@ -533,12 +972,46 @@ extern "C" {
      
      the return value is true only if a point was found
      
+     this version searches down the shortest branch first
+     
+     =======================================================================
+ 
+     =======================================================================
+     int CNearTreeLeftNearestNeighbor ( const CNearTreeHandle treehandle, 
+     const double dRadius,  
+     void CNEARTREE_FAR *  CNEARTREE_FAR * coordClosest,
+     void CNEARTREE_FAR * CNEARTREE_FAR * objClosest,   
+     const void CNEARTREE_FAR * coord )
+     
+     Function to search a Neartree for the object closest to some probe point, coord. This function
+     is only here so that the function CNearTreeNearest can be called without having dRadius const
+     
+     dRadius is the maximum search radius - any point farther than dRadius from the probe
+     point will be ignored
+     
+     coordClosest is the coordinate vector into which the coordinates of the nearest point
+     will be stored
+     
+     objClosest is the address into which a pointer to the object associated with coordClosest
+     will be stored
+     
+     coord  is the probe point
+     
+     the return value is true only if a point was found
+     
+     this version searches down the left branch first
+     
      =======================================================================
      */
     
     int CNearTreeNearestNeighbor (const CNearTreeHandle treehandle, 
                                   const double dRadius,  
-                                  void CNEARTREE_FAR *  CNEARTREE_FAR * coordClosest,
+                                  void CNEARTREE_FAR * CNEARTREE_FAR * coordClosest,
+                                  void CNEARTREE_FAR * CNEARTREE_FAR * objClosest, 
+                                  const void CNEARTREE_FAR * coord );
+    int CNearTreeLeftNearestNeighbor (const CNearTreeHandle treehandle, 
+                                  const double dRadius,  
+                                  void CNEARTREE_FAR * CNEARTREE_FAR * coordClosest,
                                   void CNEARTREE_FAR * CNEARTREE_FAR * objClosest, 
                                   const void CNEARTREE_FAR * coord );
     /*
@@ -779,13 +1252,48 @@ extern "C" {
      
      the return value is 0 only if a point was found within dRadius
      
+     this version search down the shortest branch first
+     
      =======================================================================
+     =======================================================================
+     
+     int CNearTreeLeftNearest ( const CNearTreeHandle treehandle, 
+     double CNEARTREE_FAR *dRadius,  
+     void CNEARTREE_FAR * CNEARTREE_FAR * coordClosest,
+     void CNEARTREE_FAR * CNEARTREE_FAR * objClosest,
+     const void CNEARTREE_FAR * coord )
+     
+     Function to search a Neartree for the object closest to some probe point, coord.
+     This function is called by CNearTreeNearestNeighbor.
+     
+     *dRadius is the smallest currently known distance of an object from the probe point
+     and is modified as the search progresses.
+     
+     coordClosest is the returned closest point
+     to the probe point that can be found in the Neartree
+     
+     objClosest is a pointer to a pointer to hold the corresponding object or is NULL
+     
+     coord  is the probe point
+     
+     the return value is 0 only if a point was found within dRadius
+     
+     this version searches down the left branch first
+     
+     =======================================================================
+     
      */
     int CNearTreeNearest ( const CNearTreeHandle treehandle, 
                           double CNEARTREE_FAR * dRadius,  
                           void CNEARTREE_FAR * CNEARTREE_FAR * coordClosest,
                           void CNEARTREE_FAR * CNEARTREE_FAR * objClosest,
                           const void CNEARTREE_FAR * coord );
+
+    int CNearTreeLeftNearest ( const CNearTreeHandle treehandle, 
+                          double CNEARTREE_FAR * dRadius,  
+                          void CNEARTREE_FAR * CNEARTREE_FAR * coordClosest,
+                          void CNEARTREE_FAR * CNEARTREE_FAR * objClosest,
+                          const void CNEARTREE_FAR * coord );
     /*
      =======================================================================
      int CNearTreeFindFarthest ( const CNearTreeHandle treehandle,
